// os9.c

// BVDP 98/05/28
// Read os9 disks from a image on the HD

#include <stdio.h>
#include <stddef.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <getopt.h>
#include <unistd.h>

#define int4(a) ( (*a<<24) + (*(a+1)<<16) + (*(a+2)<<8) + *(a+3) )
#define int3(a) ( (*a<<16) + (*(a+1)<<8) + *(a+2) )
#define int2(a) ( (*a<<8) + *(a+1) )

typedef struct
{
	u_char	lsn[3],
		num[2];
}SEG;

typedef struct 
{
	u_char	fd_att,
		fd_own[2],
		fd_dat[5],
		fd_lnk,
		fd_siz[4],
		fd_creat[3];
	SEG	fd_seg[48];
}INODE;

typedef struct 
{
	u_char	name[28],
		res[1],
		lsn[3];
}DENTRY;

struct lsn0
{
	u_char	dd_tot[3],
		dd_tks[1],
		dd_map[2],
		dd_bit[2],
		dd_dir[3],
		dd_own[2],
		dd_att[1],
		dd_dsk[2],
		dd_fmt[1],
		dd_spt[2],
		dd_res[2],
		dd_bt[3],
		dd_bsz[2],
		dd_dat[5],
		dd_nam[32],
		dd_opt[32],
		dd_res1[1],
		dd_sync[4];
}idsect;


FILE *fd;
int total_sectors;
int sect_size;
int root_lsn;
char debug;


//////// function prototypes

char * dirname(u_char *f);
int fsize(FILE *);
int isdir(DENTRY *d);
void do_dir(int lsn, int opt);
void error(char *msg, ...);
void extract_all(int lsn);
void get_inode(int node, INODE *);
void os9read(u_char *buff, int start, int count, INODE *node);
void print_dir_entry(DENTRY *d);
void print_dir_entry_long(DENTRY *d);
void read_lsn0(void);
void usage(void);

/////////////////////////////////////////////////////////////////////

int
main(int argc, char *argv[] )
{
	int t;
	int opt='s';
	char *outdir=NULL;
	
	while((t=getopt(argc, argv, "xlLd:sh?v"))!=-1)
	{
		switch(t)
		{
			case 'd':
				if(optarg==NULL) error("Need dirname with -d option");
				outdir=optarg;
				break;
				
			case 's': 		// short dir (default)
			case 'l':		// long dir
			case 'L':		// very long dir
			case 'i':		// info 
			case 'x':		// extract all files
				opt=t;
				break;
			
			case 'v':
				debug=1;
				break;
				
			case 'h':		// help 
			case '?':
			default: usage();
		}
	}

	if(optind>=argc) error("You must use the name of one OS9 image file. Try os9 -h for a useful message");
	
	fd=fopen(argv[optind], "r");
	if(fd==NULL) error("Can't open %s for reading", argv[optind]);
	
	if(outdir)
	{
		mkdir(outdir, S_IRWXU);
		if(chdir(outdir) ) error("Can't chd to %s", outdir);
	}

	
	read_lsn0();
	
	if( total_sectors*sect_size != fsize(fd) )
		error("According to LSN0 of the image file there are %d %d byte sectors.\nThe file size should be %d, but it is %d.\nLooks like the image file is incorrect\n", total_sectors, sect_size, total_sectors*sect_size, fsize(fd) );
	
	
	
	switch(opt)
	{
		case 's':
		case 'L':
		case 'l':
			printf("Disk name: %s\n", dirname(idsect.dd_nam) );
			do_dir(root_lsn, opt);
			break;
			
		case 'x':
			extract_all(root_lsn);
			break;
	}
	
	return 0;
}

/////////////////////////////////////////////
// read bytes into buffer from a file

void
os9read(u_char *buff, int start, int count, INODE *node)
{
	int offset;
	int sz, st, n;
	SEG *s;
	int t;
	int done;
	
	s=node->fd_seg;
	
	if(debug) fprintf(stderr, "%s: buff=%x startoff=%x count=%x\n",
		__FUNCTION__, buff, start, count);
	
	for(t=0, done=0, st=0; done<count; t++)
	{
		if(debug) fprintf(stderr, "   Testing lsn %x image offset=%x\n",
			int3(s[t].lsn), int3(s[t].lsn)*sect_size);
		
		offset=int3(s[t].lsn)*sect_size;
		sz=int2(s[t].num)*sect_size;
		
		if(st>=start)
		{
			fseek(fd, offset, SEEK_SET);
			n=sz;
			if(n>count-done) n=count-done;
			
			if(debug) fprintf(stderr, "   Getting %x bytes\n", n);
			
			fread(buff, 1, n, fd);
			buff+=n;
			count-=n;
		}
		
		st+=sz;
	}
}

////////////////////////////////////////////////////////
// Extract a specified file. Need name and starting lsn

void
extract(char *name, int lsn)
{
	INODE node;
	int size;
	u_char *buff;
	FILE *fd;
	
	if(debug) fprintf(stderr, "%s: File=%s lsn=%x\n", __FUNCTION__, name, lsn);
	
	get_inode(lsn, &node);
	size=int4(node.fd_siz);
	buff=malloc(size);
	if(buff==NULL) error("Can't allocate memory for file %s, size %d", name, size);
	
	os9read(buff, 0, size, &node);
	
	fd=fopen(name, "w");
	if(fd==NULL) error("Can't create new file %s", name);
	fwrite(buff, 1, size, fd);
	fclose(fd);
}

///////////////////////////////////////////////////////
// Extract all files. Really need to add a feature so
// that only specified files are grabbed.

// This starts at the root directory and goes though each file.
// If a file is a directory, the directory is created, the program
// cd()s to the new directory and does a recursive call here.

void
extract_all(int lsn)
{
	INODE lsd;
	DENTRY  *dbuff;
	int fsize, t, num;
	
	get_inode(lsn, &lsd);
	fsize=int4(lsd.fd_siz);
	
	dbuff=malloc(fsize);
	os9read((u_char*)dbuff, 0, fsize, &lsd);
	
	num=fsize/sizeof(DENTRY);
	
	for(t=2; t<num; t++)
	{
		if(dbuff[t].name[0]==0) continue;
		
		if(debug) fprintf(stderr, "%s: %s isdir=%d\n",
			__FUNCTION__, dirname(dbuff[t].name), isdir(&dbuff[t]) );
		
		if(isdir(&dbuff[t]))
		{
			mkdir(dirname(dbuff[t].name), S_IRWXU);
			chdir(dirname(dbuff[t].name) );
			extract_all( int3(dbuff[t].lsn) );
			chdir("..");
		}
		else extract(dirname(dbuff[t].name), int3(dbuff[t].lsn) );
	}
}

///////////////////////////////////////////////////////////////
// Display a directory.
// Valid opts are:
//     L - very long format (really only for debugging)
//     l - long format (like dir -x)
//     any other value does a short, 2 column listing

void
do_dir(int lsn, int opt)
{
	INODE lsd;
	DENTRY  *dbuff;
	int fsize, t, n, num;
	
	get_inode(lsn, &lsd);
	fsize=int4(lsd.fd_siz);
	
	dbuff=malloc(fsize);
	os9read((u_char*)dbuff, 0, fsize, &lsd);
	
	num=fsize/sizeof(DENTRY);
	
	for(t= ( opt=='L' ? 0 : 2), n=0; t<num;  t++)
	{
		if(dbuff[t].name[0]==0) continue;
		
		if(opt=='l') print_dir_entry(&dbuff[t]);
		else if(opt=='L') print_dir_entry_long(&dbuff[t]);
		else
		{
			printf("%-32s", dirname(dbuff[t].name) );
			if(++n%2) printf("\n");
			
		}
	}
	
	printf("\n");

	for(t=2; t<num; t++)
	{
		if(isdir(&dbuff[t]))
		{
			printf("%s:\n", dirname(dbuff[t].name) );
			do_dir( int3(dbuff[t].lsn), opt);
		}
	}
}

/////////////////////////////////////////////
// return 1 if this is a DIR, 0 if file

int
isdir(DENTRY *d)
{
	INODE node;
	
	get_inode(int3(d->lsn), &node);
	return (node.fd_att & 0x80);
}

////////////////////////////////////////////////////
// display a directory entry, normal long format

void
print_dir_entry(DENTRY *d)
{
	INODE node;
	static char *atts="DsEWRewr";
	int c, t;
	
	get_inode(int3(d->lsn), &node);
	
	for(t=node.fd_att, c=0; c<8; t=t<<1, c++)
	{
		if(t & 0x80) printf("%c", atts[c]);
		else printf("-");
	}
	
	printf("   %04d/%02d/%02d %02d:%02d", node.fd_dat[0]+1900, node.fd_dat[1], node.fd_dat[2], node.fd_dat[3], node.fd_dat[4]);
	printf("  %7d  ", int4(node.fd_siz) );
	
	printf("%s\n", dirname(d->name) );
}


////////////////////////////////////////////////////
// display a directory entry, extra-long format
// This included create date and lsn info

void
print_dir_entry_long(DENTRY *d)
{
	INODE node;
	static char *atts="DsEWRewr";
	int c, t;
	
	get_inode(int3(d->lsn), &node);
	
	for(t=node.fd_att, c=0; c<8; t=t<<1, c++)
	{
		if(t & 0x80) printf("%c", atts[c]);
		else printf("-");
	}
	
	printf("   %04d/%02d/%02d %02d:%02d   %02d/%02d/%02d", node.fd_dat[0]+1900, node.fd_dat[1], node.fd_dat[2], node.fd_dat[3], node.fd_dat[4], node.fd_creat[0], node.fd_creat[1], node.fd_creat[2]);
	printf("  %7d  ", int4(node.fd_siz) );
	printf("  %6x  ", int3(d->lsn) );
	printf("%s\n", dirname(d->name) );
}

//////////////////////////////////////////////////
// Convert a os9 filename to a string. OS9 stores
// filename with the last character in the name as
// with bit 7 set. This is also used in the lsn0 for
// the disk name. I seem to recall that the name in
// lsn0 can be 0 terminated, so that is also checked

char *
dirname(u_char *f)
{
	static char name[32];
	char  *s;
	
	for(s=name; ; )
	{
		if(*f==0) break;	// probably never happens
		if(*f>0x7f)
		{
			*s++=*f-0x80;
			*s=0;
			break;
		}
		*s++=*f++;
	}
	return name;
}

////////////////////////////////////////////////////////
// Read an inode. Pass the lsn of the node

void
get_inode(int n, INODE *inode)
{
	fseek(fd, n*sect_size, SEEK_SET);
	fread(inode, 1, sizeof(INODE), fd);	
}

///////////////////////////////////////////////
// get some info from lsn0 and save it.
// lots more there than we need...

void
read_lsn0()
{
	fseek(fd, 0, SEEK_SET);
	fread(&idsect, 1, sizeof(struct lsn0), fd);
	
	total_sectors=int3(idsect.dd_tot);
	sect_size=idsect.dd_bit[1] * 256;
	root_lsn=int3(idsect.dd_dir);
}

///////////////////////////////////////
// get the size of a open file. This is 
// portable??

int
fsize(FILE *fd)
{
	fseek(fd, 0, SEEK_END);
	return ftell(fd);
}


///////////////////////////////////////
// Display an error message and exit

void
error(char *msg, ...)
{
	va_list ap;
	
	if(msg)
	{
		va_start(ap, msg);
		vfprintf(stderr, msg, ap);
		va_end(ap);
	}
	
	fprintf(stderr, "\nos9 terminated\n");
	exit(1);
}

//////////////////////////////////////////////
// Display a usage message and exit

void
usage(void)
{
	int t;
	static char *msg[]=
	{
		"\nUsage: os9disk [options] imagefile",
		"Function: extracts files and directories from os9 image files",
		"Options:",
		"  -s       display short directory (default)",
		"  -d DIR   set destination directory to DIR (default=.)",
		"  -l       display long directory",
		"  -L       long dir with FDAs && create date",
		"  -x       extract all files",
		"  -v       print debug info to stdout",
		"",
		"Image files are created by running a fdrawcmd script to copy a OS9",
		"disk to your hard drive. The file README.os9 has more information",
		"",
		"Another fine program from Bob van der Poel Software <bvdpoel@uniserve.com>",
		""
	};
	
	for(t=0; t<sizeof(msg)/sizeof(msg[0]); fprintf(stderr, "%s\n", msg[t++]));
	exit(1);
}
